import re

from .CatsParseException import CatsParseException

REGEXES = {
    'user_type_name': re.compile(r'^[A-Z][A-Z0-9]*[a-z][a-zA-Z0-9]*$'),  # UpperCamelCase
    'const_property_name': re.compile(r'^[A-Z_][A-Z0-9_]*$'),  # TITLE_CASE
    'property_name': re.compile(r'^[a-z_][a-z0-9_]*$'),  # snake_case

    'int_or_uint': re.compile(r'^(?P<unsigned>u)?int(?P<size>8|16|32|64)$'),
    'binary_fixed_type': re.compile(r'^binary_fixed\((?P<size>0x[0-9A-F]+|[0-9]+)\)$'),
    'dec_or_hex': re.compile(r'^(0x[0-9A-F]+|[0-9]+)$'),
}


def _match_regex_or_throw(regex_key, line):
    match = REGEXES[regex_key].match(line)
    if not match:
        raise CatsParseException('unable to parse "{}": {}'.format(regex_key, line))

    return match


class TypeNameChecker:
    @staticmethod
    def require_user_type(type_name):
        """Raises an exception if the specified name is not a valid user type name"""
        _match_regex_or_throw('user_type_name', type_name)
        return type_name

    @staticmethod
    def require_const_property(type_name):
        """Raises an exception if the specified name is not a valid const property name"""
        _match_regex_or_throw('const_property_name', type_name)
        return type_name

    @staticmethod
    def require_property(type_name):
        """Raises an exception if the specified name is not a valid property name"""
        if REGEXES['int_or_uint'].match(type_name):
            raise CatsParseException('unable to parse "property_name": {}'.format(type_name))

        _match_regex_or_throw('property_name', type_name)
        return type_name

    @staticmethod
    def require_primitive(type_name):
        """Raises an exception if the specified name is not a valid primitive name"""
        _match_regex_or_throw('int_or_uint', type_name)
        return type_name


def is_primitive(type_name):
    """Returns true if the specified name is a valid primitive name"""
    return REGEXES['int_or_uint'].match(type_name)


def is_dec_or_hex(string):
    """Returns true if the specified string is a valid decimal or hexidecimal number"""
    return REGEXES['dec_or_hex'].match(string)


def parse_dec_or_hex(string):
    """Parses a string as either decimal or hexidecimal"""
    base = 16 if string.startswith('0x') else 10
    return int(string, base)


def is_builtin(type_name):
    return REGEXES['int_or_uint'].match(type_name) or REGEXES['binary_fixed_type'].match(type_name)


def parse_builtin(type_name):
    """Parses a builtin type, either binary_fixed or a uint alias"""
    is_unsigned = True
    binary_fixed_type_match = REGEXES['binary_fixed_type'].match(type_name)
    if binary_fixed_type_match:
        type_descriptor = {'size': parse_dec_or_hex(binary_fixed_type_match.group('size'))}
    else:
        match = _match_regex_or_throw('int_or_uint', type_name)
        is_unsigned = bool(match.group('unsigned'))
        uint_byte_count = int(match.group('size')) // 8
        type_descriptor = {'size': uint_byte_count}

    return {**type_descriptor, 'type': 'byte', 'signedness': 'unsigned' if is_unsigned else 'signed'}



from .CatsParseException import CatsParseException


class Scope:
    """Tuple composed of filename and line number"""
    def __init__(self, name):
        self.name = name
        self.line_number = 0


class ScopeManager:
    """Manages the current scope composed of filename and line number"""
    def __init__(self):
        self.scopes = [Scope('<unknown>')]

    def push_scope(self, name):
        """Pushes the input scope"""
        self.scopes.append(Scope(name))

    def scope(self):
        """Gets the current location"""
        return ['{}:{}'.format(scope.name, scope.line_number) for scope in self.scopes][::-1]

    def pop_scope(self):
        """Pops the input scope"""
        if 1 == len(self.scopes):
            raise CatsParseException('CatsParser cannot pop default scope')

        self.scopes.pop()

    def increment_line_number(self):
        """Increments the line number for the current scope"""
        self.scopes[-1].line_number += 1



from .RegexParserFactory import RegexParserFactory


class ImportResult:
    """Information about an import statement"""
    def __init__(self, import_file):
        self.import_file = import_file

    def __eq__(self, rhs):
        return isinstance(rhs, ImportResult) and self.import_file == rhs.import_file


class ImportParser:
    """Parser for `import` statements"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)
        return ImportResult(match.group('filename'))


class ImportParserFactory(RegexParserFactory):
    """Factory for creating import parsers"""
    def __init__(self):
        super().__init__(r'import "(?P<filename>[\S ]+)"', ImportParser)



import argparse
import glob
import importlib
import os
import re
import sys

import yaml

from .CatsParseException import CatsParseException
from .CatsParser import CatsParser


class MultiFileParser:
    """CATS parser that resolves imports in global namespace"""
    def __init__(self):
        self.cats_parser = CatsParser(self._process_import_file)
        self.dirname = None
        self.imported_filenames = []

    def set_include_path(self, include_path):
        self.dirname = include_path

    def parse(self, schema_filename):
        self._process_file(schema_filename)

    def _process_import_file(self, filename):
        if filename in self.imported_filenames:
            return

        self.imported_filenames.append(filename)
        filename = os.path.join(self.dirname, filename)
        self._process_file(filename)

    def _process_file(self, filename):
        self.cats_parser.push_scope(filename)

        with open(filename) as input_file:
            lines = input_file.readlines()
            for line in lines:
                self.cats_parser.process_line(line)

        self.cats_parser.commit()
        self.cats_parser.pop_scope()


def _generate_output(generator_class, output_path, schema, options):
    os.makedirs(output_path, exist_ok=True)
    generator = generator_class(schema, options)
    for generated_descriptor in generator:
        output_filename = os.path.join(output_path, generated_descriptor.filename)
        with open(output_filename, 'w', newline='\n') as output_file:
            for line in generated_descriptor.code:
                output_file.write('%s\n' % line)


def _build_available_generator_map():
    generator_map = {}
    if os.path.exists('generators'):
        generator_file_name_pattern = re.compile(r'^generators/(?P<short_name>.*)/(?:.+File|Builder)Generator\.py$', re.IGNORECASE)

        for generator_path in glob.glob('generators/*/*Generator.py'):
            generator_file_name_match = generator_file_name_pattern.match(generator_path)
            if generator_file_name_match:
                generator_class_name = generator_path.replace('/', '.')[:-3]
                generator_map[generator_file_name_match.group('short_name')] = generator_class_name

    if generator_map:
        print('autodetected the following generators:')
        for short_name in generator_map:
            print(' + {} => {}'.format(short_name, generator_map[short_name]))

        print()

    return generator_map


def _dump_type_descriptors(type_descriptors, out):
    needs_separator = False
    for key in type_descriptors:
        if needs_separator:
            out.write('\n')

        out.write('name: {}\n'.format(key))
        yaml.dump(type_descriptors[key], out)
        needs_separator = True


def main():
    generator_map = _build_available_generator_map()
    available_generator_names = generator_map.keys() if generator_map else None

    parser = argparse.ArgumentParser(
        prog=None if globals().get('__spec__') is None else 'python -m {}'.format(__spec__.name.partition('.')[0]),
        description='CATS code generator'
    )
    parser.add_argument('-s', '--schema', help='input CATS file', required=True)
    parser.add_argument('-o', '--output', help='output directory, if not provided, _generated/{generator} is used')
    parser.add_argument('-i', '--include', help='schema root directory', default='./schemas')
    parser.add_argument('-x', '--export', help='export schemas as yaml')

    parser.add_argument('-g', '--generator', help='generator to use to produce output files', choices=available_generator_names)
    parser.add_argument('-c', '--copyright', help='file containing copyright data to use with output files', default='../HEADER.inc')
    args = parser.parse_args()

    file_parser = MultiFileParser()
    file_parser.set_include_path(args.include)

    try:
        file_parser.parse(args.schema)
        type_descriptors = file_parser.cats_parser.type_descriptors()
    except CatsParseException as ex:
        print('\033[31m{}\033[39m'.format(ex.message))
        print('\033[33m{}\033[39m'.format(ex.scope[0]))
        for location in ex.scope[1:]:
            print(' + {}'.format(location))

        sys.exit(1)

    # dump parsed type descriptors
    if not args.generator:
        _dump_type_descriptors(type_descriptors, sys.stdout)

    if args.export:
        with open(args.export, 'wt') as out:
            _dump_type_descriptors(type_descriptors, out)

    if args.generator:
        # load and run generator
        generator_full_name = generator_map.get(args.generator, args.generator)
        generator_class_name = os.path.splitext(generator_full_name)[1][1:]

        print()
        print('loading generator {} from {}'.format(generator_class_name, generator_full_name))

        generator_module = importlib.import_module(generator_full_name)
        generator_class = getattr(generator_module, generator_class_name)

        print('loaded generator: {}'.format(generator_class))
        print()

        output_path = args.output
        if output_path is None:
            output_path = os.path.join('_generated', generator_class_name)

        _generate_output(generator_class, output_path, type_descriptors, {'copyright': args.copyright})


if '__main__' == __name__:
    main()



from .parserutils import TypeNameChecker, parse_builtin
from .RegexParserFactory import RegexParserFactory


class AliasParser:
    """Parser for `using` statements"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)

        # aliases are only supported for builtin types
        return (
            TypeNameChecker.require_user_type(match.group('alias_type_name')),
            parse_builtin(match.group('aliased_type_name'))
        )


class AliasParserFactory(RegexParserFactory):
    """Factory for creating alias parsers"""
    def __init__(self):
        super().__init__(r'using (?P<alias_type_name>\S+) = (?P<aliased_type_name>\S+)', AliasParser)



import re
from collections import OrderedDict

from .AliasParser import AliasParserFactory
from .CatsParseException import CatsParseException
from .CommentParser import CommentParser
from .EnumParser import EnumParserFactory
from .ImportParser import ImportParserFactory
from .ScopeManager import ScopeManager
from .StructParser import StructParserFactory


class CatsParser(ScopeManager):
    """Parser used to parse CATS files line by line"""
    def __init__(self, import_resolver):
        super().__init__()
        self.import_resolver = import_resolver

        self.aspect_parser = CommentParser()
        self.type_parser_factories = [
            AliasParserFactory(),
            EnumParserFactory(),
            ImportParserFactory(),
            StructParserFactory()
        ]

        self.wip_type_descriptors = OrderedDict()
        self.active_parser = None

    def process_line(self, line):
        """Processes the next line of input"""
        try:
            self._process_line(line)
        except Exception as ex:
            raise CatsParseException(self.scope(), ex) from ex

    def commit(self):
        """Completes processing of current type"""
        self._close_type()

    def _process_line(self, line):
        self.increment_line_number()

        # check if current line is a cross cutting concern
        line_stripped = line.strip()
        if self.aspect_parser.try_process_line(line_stripped):
            return

        # something else, so attach current aspect state to it
        partial_descriptor = self.aspect_parser.commit()

        # ignore blank lines
        if not line_stripped:
            return

        # close a type iff an unindented non-empty line is found
        if self.active_parser and not line.startswith('\t'):
            self._close_type()

        active_factories = self.type_parser_factories if not self.active_parser else self.active_parser.factories()

        try:
            factory = next(factory for factory in active_factories if factory.is_match(line_stripped))
        except StopIteration as ex:
            raise CatsParseException('none of the parsers matched the line "{}"'.format(line_stripped)) from ex

        parser = factory.create()
        parse_result = parser.process_line(line_stripped)

        # create a new scope if the current symbol is a composite
        if not parse_result:
            self.active_parser = parser
            self.active_parser.partial_descriptor = partial_descriptor
            return

        if self.active_parser:
            if 'type' in parse_result:
                # perform extra validation on some property links for better error detection/messages
                if 'sort_key' in parse_result:
                    # sort key processing will only occur if linked field type already exists
                    self._require_type_with_field(parse_result['type'], parse_result['sort_key'])

                # perform inline struct expansion
                if self._handle_inline_struct({**parse_result, **partial_descriptor}):
                    return

            self.active_parser.append({**parse_result, **partial_descriptor})
        elif hasattr(parse_result, 'import_file'):
            self.import_resolver(parse_result.import_file)
        else:
            self._set_type_descriptor(parse_result[0], {**parse_result[1], **partial_descriptor})

    def _handle_inline_struct(self, parse_result):
        linked_type_name = parse_result['type']
        member_type_descriptor = self._require_known_type(linked_type_name)

        # special primitive types, like byte, do not have explicit type descriptors
        is_inline_struct = member_type_descriptor and 'inline' == member_type_descriptor.get('disposition')
        is_inline_member = 'inline' == parse_result.get('disposition')

        if is_inline_member:
            if not member_type_descriptor:
                raise CatsParseException('type "{}" cannot be inlined'.format(linked_type_name))
        else:
            if is_inline_struct:
                raise CatsParseException('inline struct type "{}" can only be used as a named inline'.format(linked_type_name))

            return False

        if is_inline_struct and not parse_result.get('name'):
            raise CatsParseException('inline struct type "{}" must be named when inlined'.format(linked_type_name))

        if not is_inline_struct:
            if parse_result.get('name'):
                raise CatsParseException('struct type "{}" must be unnamed when inlined'.format(linked_type_name))

            return False

        comment_map = self._build_comment_map(parse_result['comments'])

        for inline_struct_member_descriptor in member_type_descriptor['layout']:
            property_descriptor = {**inline_struct_member_descriptor}
            property_descriptor['comments'] = '\n'.join(comment_map.get(property_descriptor['name'], []))

            property_name = parse_result['name']
            property_descriptor['name'] = self._make_inlined_name(property_name, property_descriptor['name'])

            if property_descriptor.get('condition'):
                property_descriptor['condition'] = self._make_inlined_name(property_name, property_descriptor['condition'])

            if (property_descriptor.get('size')
                    and any(property_descriptor['size'] == descriptor['name'] for descriptor in member_type_descriptor['layout'])):
                property_descriptor['size'] = self._make_inlined_name(property_name, property_descriptor['size'])

            self.active_parser.append(property_descriptor)

        return True

    @staticmethod
    def _make_inlined_name(property_name_prefix, internal_property_name):
        if '__value__' == internal_property_name:
            return property_name_prefix

        return'{}_{}'.format(property_name_prefix, internal_property_name)

    @staticmethod
    def _build_comment_map(comments):
        member_comment_start_regex = re.compile(r'^\[(?P<comment_key>\S+)\] ')

        comment_map = {}
        active_comment_key = None
        for line in comments.split('\n'):
            member_comment_start_match = member_comment_start_regex.match(line)
            if member_comment_start_match:
                active_comment_key = member_comment_start_match.group('comment_key')
                comment_map[active_comment_key] = [line[len(active_comment_key) + 3:]]
            elif active_comment_key:
                comment_map[active_comment_key].append(line)

        return comment_map

    def _close_type(self):
        if not self.active_parser:
            return

        (new_type_name, new_type_descriptor) = self.active_parser.commit()

        if 'layout' in new_type_descriptor:
            new_type_layout = new_type_descriptor['layout']
            for property_type_descriptor in new_type_layout:
                if 'condition' in property_type_descriptor:
                    # when condition is being post processed here, it is known that the linked condition field is part of
                    # the struct and the linked condition type already exists

                    # look up condition type descriptor by searching active parser descriptor layout
                    condition_field_name = property_type_descriptor['condition']
                    condition_type_descriptor = next(
                        descriptor for descriptor in new_type_layout
                        if 'name' in descriptor and descriptor['name'] == condition_field_name
                    )

                    condition_type_name = condition_type_descriptor['type']
                    condition_value = property_type_descriptor['condition_value']
                    if 'byte' == condition_type_name:
                        if not isinstance(condition_value, int):
                            error_message_format = 'condition value "{}" for "{}" must be numeric'
                            raise CatsParseException(error_message_format.format(condition_value, condition_field_name))
                    else:
                        self._require_enum_type_with_value(condition_type_name, condition_value)

                if any(disposition == property_type_descriptor.get('disposition') for disposition in ['const', 'reserved']):
                    if not isinstance(property_type_descriptor['value'], int):
                        self._require_enum_type_with_value(property_type_descriptor['type'], property_type_descriptor['value'])

                if 'inline' == new_type_descriptor.get('disposition') and 'const' == property_type_descriptor.get('disposition'):
                    error_message_format = 'inline struct "{}" cannot include const field "{}"'
                    raise CatsParseException(error_message_format.format(new_type_name, property_type_descriptor['name']))

        self._set_type_descriptor(new_type_name, {**new_type_descriptor, **self.active_parser.partial_descriptor})
        self.active_parser = None

    def _require_known_type(self, type_name):
        if 'byte' == type_name:
            return None

        if type_name not in self.wip_type_descriptors:
            raise CatsParseException('no definition for linked type "{}"'.format(type_name))

        return self.wip_type_descriptors[type_name]

    def _require_type_with_field(self, type_name, field_name):
        type_descriptor = self.wip_type_descriptors[type_name]
        if not any(field_name == field['name'] for field in type_descriptor['layout']):
            raise CatsParseException('"{}" does not have field "{}"'.format(type_name, field_name))

    def _require_enum_type_with_value(self, type_name, value_name):
        enum_type_descriptor = self.wip_type_descriptors[type_name]
        if 'values' not in enum_type_descriptor:
            raise CatsParseException('linked type "{}" must be an enum type'.format(type_name))

        if not any(value_name == value['name'] for value in enum_type_descriptor['values']):
            raise CatsParseException('linked enum type "{}" does not contain value "{}"'.format(type_name, value_name))

    def _set_type_descriptor(self, type_name, type_descriptor):
        if type_name in self.wip_type_descriptors:
            raise CatsParseException('duplicate definition for type "{}"'.format(type_name))

        self.wip_type_descriptors[type_name] = type_descriptor

    @staticmethod
    def _is_inline_struct(type_descriptor):
        return 'struct' == type_descriptor.get('type') and 'inline' == type_descriptor.get('disposition')

    def type_descriptors(self):
        """Returns all parsed type descriptors"""
        self._close_type()

        # skip inline structs as they're treated as macros and expanded by the parser
        return {
            name: self.wip_type_descriptors[name] for name in self.wip_type_descriptors
            if not self._is_inline_struct(self.wip_type_descriptors[name])
        }



import re


class RegexParserFactory:
    """Base for top-level parser factories"""
    def __init__(self, regex, parser_type):
        self.regex = re.compile('^{}$'.format(regex))
        self.parser_type = parser_type

    def is_match(self, line):
        """Returns True if the line is a match for this factory's parser"""
        return self.regex.match(line)

    def create(self):
        """Creates a new parser"""
        return self.parser_type(self.regex)



class CatsParseException(Exception):
    """Exception raised when a parse error is encountered"""

    def __init__(self, args, ex=None):
        super().__init__(args)

        self.message = None
        self.scope = None

        if isinstance(args, str):
            self.message = args
        elif isinstance(args, list):
            self.scope = args

        if isinstance(ex, CatsParseException):
            self.message = ex.message



# only comments immediately adjacent to a line will appear in the output.
# consecutive lines are concatenated with <SPACE>.
# a blank line inserts a new line `\n` in the output.

class CommentParser:
    """Reusable aspect parser for comment statements"""
    def __init__(self):
        self.comments = []

    def try_process_line(self, line):
        """Processes the current line if and only if it is a comment line"""
        if line.startswith('#'):
            self.comments.append(line[1:])
            return True

        return False

    def commit(self):
        """Postprocesses and clears all captured comments"""
        comments = ''
        needs_separator = False
        for comment in self.comments:
            comment = comment.strip()

            if not comment:
                comments += '\n'
                needs_separator = False
            else:
                if needs_separator:
                    comments += ' '

                comments += comment
                needs_separator = True

        self.comments.clear()
        return {'comments': comments}



from .CatsParseException import CatsParseException
from .CompositeTypeParser import CompositeTypeParser
from .parserutils import TypeNameChecker, is_builtin, is_dec_or_hex, is_primitive, parse_builtin, parse_dec_or_hex
from .RegexParserFactory import RegexParserFactory

# region StructParser(Factory)


class StructParser(CompositeTypeParser):
    """Parser for `struct` statements"""
    def __init__(self, regex):
        super().__init__(regex, [
            StructConstParserFactory(),
            StructInlineParserFactory(),
            StructScalarMemberParserFactory(),
            StructArrayMemberParserFactory()
        ])

    def process_line(self, line):
        match = self.regex.match(line)
        self.type_name = TypeNameChecker.require_user_type(match.group('struct_type_name'))
        self.type_descriptor = {'type': 'struct', 'layout': []}

        if match.group('inline'):
            self.type_descriptor['disposition'] = 'inline'

    def append(self, property_type_descriptor):
        self._require_no_array_with_fill_disposition()

        if 'size' in property_type_descriptor:
            self._require_known_property(property_type_descriptor['size'])

        descriptor_uid = self._get_descriptor_uid(property_type_descriptor)
        if descriptor_uid[0]:
            self._require_unknown_property(descriptor_uid)

        self.type_descriptor['layout'].append(property_type_descriptor)

    def commit(self):
        for property_type_descriptor in self.type_descriptor['layout']:
            if 'condition' in property_type_descriptor:
                self._require_known_property(property_type_descriptor['condition'], False)

        return super().commit()

    def _require_no_array_with_fill_disposition(self):
        layout = self.type_descriptor['layout']
        if not layout:
            return

        last_property = layout[-1]
        if 'fill' == last_property.get('disposition'):
            raise CatsParseException('array property with fill disposition "{}" must be last property'.format(last_property['name']))

    def _require_known_property(self, property_name, allow_numeric=True):
        # size can be a constant represented by a numeric type
        if allow_numeric and not isinstance(property_name, str):
            return

        if all('name' not in property_type_descriptor or property_name != property_type_descriptor['name']
               for property_type_descriptor in self.type_descriptor['layout']):
            raise CatsParseException('no definition for referenced property "{}"'.format(property_name))

    def _require_unknown_property(self, descriptor_uid):
        if any(descriptor_uid == self._get_descriptor_uid(property_type_descriptor)
               for property_type_descriptor in self.type_descriptor['layout']):
            raise CatsParseException('duplicate definition for property "{}"'.format(descriptor_uid))

    @staticmethod
    def _get_descriptor_uid(descriptor):
        return (descriptor.get('name'), descriptor.get('disposition'))


class StructParserFactory(RegexParserFactory):
    """Factory for creating struct parsers"""
    def __init__(self):
        super().__init__(r'(?P<inline>inline )?struct (?P<struct_type_name>\S+)', StructParser)


# endregion

# region StructConstParser(Factory)

class StructConstParser:
    """Parser for const struct members"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)
        type_name = match.group('type_name')

        disposition = match.group('disposition')
        property_name_checker = TypeNameChecker.require_const_property if 'const' == disposition else TypeNameChecker.require_property
        const_descriptor = {
            'name': property_name_checker(match.group('name')),
            'disposition': disposition,
            'value': match.group('value')
        }

        is_numeric = False
        if is_primitive(type_name):
            const_descriptor = {**const_descriptor, **parse_builtin(type_name)}
            is_numeric = True
        else:
            const_descriptor['type'] = TypeNameChecker.require_user_type(type_name)
            is_numeric = is_dec_or_hex(const_descriptor['value'])

        if is_numeric:
            const_descriptor['value'] = parse_dec_or_hex(const_descriptor['value'])

        return const_descriptor


class StructConstParserFactory(RegexParserFactory):
    """Factory for creating struct const parsers"""
    def __init__(self):
        super().__init__(r'(?P<name>\S+) = make_(?P<disposition>const|reserved)\((?P<type_name>\S+), (?P<value>\S+)\)', StructConstParser)


# endregion

# region StructInlineParser(Factory)

class StructInlineParser:
    """Parser for inline struct members"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        # type is resolved to exist upstream, so its naming doesn't need to be checked here
        match = self.regex.match(line)
        inline_descriptor = {'type': match.group('inline_type_name'), 'disposition': 'inline'}

        inline_name_prefix = match.group('name')
        if inline_name_prefix:
            inline_descriptor['name'] = inline_name_prefix

        return inline_descriptor


class StructInlineParserFactory(RegexParserFactory):
    """Factory for creating struct inline parsers"""
    def __init__(self):
        super().__init__(r'(?:(?P<name>\S+) = )?inline (?P<inline_type_name>\S+)', StructInlineParser)


# endregion

# region StructScalarMemberParser(Factory)

OPTIONAL_CONDITIONAL_REGEX_PATTERN = \
    r'(?: if (?P<condition_value>\S+) (?P<condition_negate>not )?(?P<condition_operation>equals|in) (?P<condition>\S+))?'


def parse_conditional_instruction(property_type_descriptor, match):
    if not match.group('condition_value'):
        return

    is_negated = 'not ' == match.group('condition_negate')

    property_type_descriptor['condition'] = match.group('condition')
    property_type_descriptor['condition_operation'] = ('not ' if is_negated else '') + match.group('condition_operation')

    condition_value = match.group('condition_value')
    if is_dec_or_hex(condition_value):
        condition_value = parse_dec_or_hex(condition_value)

    property_type_descriptor['condition_value'] = condition_value


class StructScalarMemberParser:
    """Parser for non-inline scalar struct members"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)
        linked_type_name = match.group('type_name')

        # type is resolved to exist upstream, so its naming doesn't need to be checked here
        if is_builtin(linked_type_name):
            property_type_descriptor = parse_builtin(linked_type_name)  # reduce builtins to byte
        else:
            property_type_descriptor = {'type': linked_type_name}

        parse_conditional_instruction(property_type_descriptor, match)

        property_type_descriptor['name'] = TypeNameChecker.require_property(match.group('name'))
        return property_type_descriptor


class StructScalarMemberParserFactory(RegexParserFactory):
    """Factory for creating struct scalar member parsers"""
    def __init__(self):
        super().__init__(
            r'(?P<name>\S+) = (?P<type_name>\S+)' + OPTIONAL_CONDITIONAL_REGEX_PATTERN,
            StructScalarMemberParser)


# endregion

# region StructArrayMemberParser(Factory)

class StructArrayMemberParser:
    """Parser for non-inline array struct members"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)

        # type is resolved to exist upstream, so its naming doesn't need to be checked here
        element_type_name = match.group('element_type_name')

        if 'byte' == element_type_name:
            raise CatsParseException('explicit use of byte is deprecated please use uint8 or int8 instead')

        property_type_descriptor = {}
        if is_primitive(element_type_name):
            builtin_type_descriptor = parse_builtin(element_type_name)
            element_type_name = builtin_type_descriptor['type']

            property_type_descriptor['element_disposition'] = {**builtin_type_descriptor}
            del property_type_descriptor['element_disposition']['type']

        property_type_descriptor['type'] = element_type_name

        # size can be interpreted in different ways for count-based arrays
        # size must be a field reference for size-based arrays
        array_size = match.group('size')
        if not match.group('byte_size'):
            if is_dec_or_hex(array_size):
                array_size = parse_dec_or_hex(array_size)

            if '__FILL__' == array_size:
                array_size = 0
                property_type_descriptor['disposition'] = 'array fill'  # expands to fill the rest of the structure
            else:
                property_type_descriptor['disposition'] = 'array'
        else:
            property_type_descriptor['disposition'] = 'array sized'  # fits a predefined size

        property_type_descriptor['size'] = array_size

        if match.group('sort_key'):
            property_type_descriptor['sort_key'] = match.group('sort_key')

        parse_conditional_instruction(property_type_descriptor, match)

        property_type_descriptor['name'] = TypeNameChecker.require_property(match.group('name'))
        return property_type_descriptor


class StructArrayMemberParserFactory(RegexParserFactory):
    """Factory for creating struct array member parsers"""
    def __init__(self):
        super().__init__(
            r'(?P<name>\S+) = array\((?P<element_type_name>\S+), (?P<byte_size>size=)?(?P<size>\S+)(?:, sort_key=(?P<sort_key>\S+))?\)'
            + OPTIONAL_CONDITIONAL_REGEX_PATTERN,
            StructArrayMemberParser)

# endregion



from .CatsParseException import CatsParseException
from .CompositeTypeParser import CompositeTypeParser
from .parserutils import TypeNameChecker, parse_builtin, parse_dec_or_hex
from .RegexParserFactory import RegexParserFactory


class EnumParser(CompositeTypeParser):
    """Parser for `enum` statements"""
    def __init__(self, regex):
        super().__init__(regex, [EnumValueParserFactory()])

    def process_line(self, line):
        match = self.regex.match(line)
        self.type_name = TypeNameChecker.require_user_type(match.group('enum_type_name'))

        base_type = TypeNameChecker.require_primitive(match.group('underlying_type_name'))
        builtin_type_descriptor = parse_builtin(base_type)
        self.type_descriptor = {
            'type': 'enum',
            'size': builtin_type_descriptor['size'],
            'signedness': builtin_type_descriptor['signedness'],
            'values': []
        }

    def append(self, property_value_descriptor):
        self._require_unknown_property(property_value_descriptor['name'])

        self.type_descriptor['values'].append(property_value_descriptor)

    def _require_unknown_property(self, property_name):
        if any(property_name == property_type_descriptor['name'] for property_type_descriptor in self.type_descriptor['values']):
            raise CatsParseException('duplicate definition for enum value "{}"'.format(property_name))


class EnumParserFactory(RegexParserFactory):
    """Factory for creating enum parsers"""
    def __init__(self):
        super().__init__(r'enum (?P<enum_type_name>\S+) : (?P<underlying_type_name>u?int\d+)', EnumParser)


class EnumValueParser:
    """Parser for enum values"""
    def __init__(self, regex):
        self.regex = regex

    def process_line(self, line):
        match = self.regex.match(line)
        return {
            'name': TypeNameChecker.require_const_property(match.group('name')),
            'value': parse_dec_or_hex(match.group('value'))
        }


class EnumValueParserFactory(RegexParserFactory):
    """Factory for creating enum value parsers"""
    def __init__(self):
        super().__init__(r'(?P<name>\S+) = (?P<value>\S+)', EnumValueParser)



from setuptools import find_packages, setup

with open('README.md', 'r') as readme_file:
    README = readme_file.read()

NAME = 'catparser'

with open('version.txt', 'r') as version_file:
    VERSION = version_file.read().strip()

REQUIRES = [
]

setup(
    name=NAME,
    version=VERSION,
    description='Symbol Catbuffer Parser',
    author='NEM Group',
    author_email='dev@nem.software',
    url='https://github.com/symbol/catbuffer-parser',
    keywords=['symbol', 'catbuffer', 'catparser', 'parser', 'catbuffer-parser'],
    install_requires=REQUIRES,
    packages=find_packages(exclude=('tests', 'tests/*')),
    include_package_data=True,
    obsoletes=['catbuffer_parser'],
    license='MIT',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ]
)



class CompositeTypeParser:
    """Base for composite type parsers"""
    def __init__(self, regex, factories):
        self.regex = regex
        self.sub_factories = factories
        self.type_name = None
        self.type_descriptor = None

    def factories(self):
        """Gets sub-parsers for this composite type parser"""
        return self.sub_factories

    def commit(self):
        """Returns the composite type tuple"""
        return (self.type_name, self.type_descriptor)
